package org.mortbay.cometd.client;

import java.net.InetSocketAddress;
import java.security.SecureRandom;
import java.util.Map;
import java.util.concurrent.atomic.AtomicInteger;

import org.mortbay.jetty.client.HttpClient;
import org.mortbay.jetty.client.HttpDestination;
import org.mortbay.thread.BoundedThreadPool;
import org.mortbay.util.ajax.JSON;

import dojox.cometd.Bayeux;
import dojox.cometd.Client;
import dojox.cometd.Listener;

public class BayeuxLoadGenerator
{
    SecureRandom _random= new SecureRandom();
    HttpClient http;
    InetSocketAddress address=new InetSocketAddress("localhost",8080);
    String uri="/cometd";
    int clients=1000;
    BayeuxClient[] client;
    int _usersPerRoom=10;
    int _burst=20;
    long _pauseConnect=30;
    long _pause=100;
    int _chat=5000;
    long _maxLatency;
    long _totalLatency;
    int _got;
    AtomicInteger _subscribed = new AtomicInteger();
    
    public BayeuxLoadGenerator() throws Exception
    {
        http=new HttpClient();
        
        http.setConnectorType(HttpClient.CONNECTOR_SELECT_CHANNEL);
        // http.setConnectorType(HttpClient.CONNECTOR_SOCKET);
        http.setMaxConnectionsPerAddress(10000);
        
        BoundedThreadPool pool = new BoundedThreadPool();
        pool.setMaxThreads(255);
        pool.setDaemon(true);
        http.setThreadPool(pool);
        http.start();
       
        
    }
    
    
    public void generateLoad() throws Exception
    {
        client = new BayeuxClient[clients];

        HttpDestination destination = http.getDestination(address,false);
        
        for (int i=0;i<clients;i++)
        {
            client[i] = new BayeuxClient(http,address,uri)
            {

                public void deliver(Client from, Map<String, Object> message)
                {
                    if (Bayeux.META_SUBSCRIBE.equals(message.get(Bayeux.CHANNEL_FIELD)) &&
                        ((Boolean)message.get(Bayeux.SUCCESSFUL_FIELD)).booleanValue())
                        _subscribed.incrementAndGet();
                    super.deliver(from,message);
                }
            };
            
            int room=i/_usersPerRoom;
            final String name = "Client"+(i>=1000?"":i>=100?"0":i>=10?"00":"000")+i;
            final String id=client[i].getClientId();
            System.err.println(name);
            Listener listener = new Listener()
            {
                public void deliver(Client fromClient, String toChannel, Object data, String msgId)
                {
                    if (data!=null)
                    {
                        // System.err.println(name+": "+data);
                        if (msgId!=null)
                        {
                            long latency= System.currentTimeMillis()-Long.parseLong(msgId);
                            synchronized(BayeuxLoadGenerator.this)
                            {
                                _got++;
                                if (_maxLatency<latency)
                                    _maxLatency=latency;
                                _totalLatency+=latency;
                            }
                        }
                    }
                }

                public void removed(String clientId, boolean timeout)
                {
                }
                
                public String toString()
                {
                    return name+"/"+id;
                }
            };
            client[i].setListener(listener);
            client[i].subscribe("/chat/demo/"+room);
            client[i].start();
            Thread.sleep(_pauseConnect);
        }

        Thread.sleep(_usersPerRoom*_pause);
        
        while(_subscribed.get()<clients)
        {
            System.err.println(destination.toDetailString());
            System.err.println("Subscribed:"+_subscribed.get()+" < "+clients+" ...");
            Thread.sleep(5000);
        }
        
        long start=System.currentTimeMillis();
        for (int i=0;i<_chat;i++)
        {
            System.err.print(i);
            System.err.print(',');
            int c = _random.nextInt(clients);
            int room=c/_usersPerRoom;
            final String name = "Client"+(i>=1000?"":i>=100?"0":i>=10?"00":"000")+i;
            client[c].publish("/chat/demo/"+room,new JSON.Literal("{\"user\":\""+name+"\",\"chat\":\"How now brown cow. "+i+"\"}"),""+System.currentTimeMillis());
            
            if (i%_burst==0)
            {
                System.err.println();
                Thread.sleep(_pause);
            }
        }

        Thread.sleep(_maxLatency);

        for (int i=0;i<clients;i++)
        {
            if (!client[i].isPolling())
                System.err.println("PROBLEM WITH "+client[i]);
        }

        System.err.println();
        
        while (_got<_chat*_usersPerRoom) 
        {
            System.err.println(destination.toDetailString());
            System.err.println("Got:"+_got+" < "+(_chat*_usersPerRoom)+" ...");
            
            Thread.sleep(1000);
        }
        
        for (int i=0;i<clients;i++)
        {
            client[i].remove(false);
        }

        long end=System.currentTimeMillis();
        
        System.err.println("Got "+_got+" at "+(_got*1000/(end-start))+"/s, ave latency="+(_totalLatency/_got)+"ms, max latency="+_maxLatency+"ms");

    }
    
    
    public static void main(String[] args) throws Exception
    {
        BayeuxLoadGenerator gen = new BayeuxLoadGenerator();
        
        if (args.length>0)
            gen.clients=Integer.parseInt(args[0]);
        if (args.length>1)
            gen._chat=Integer.parseInt(args[1]);
            
        
        gen.generateLoad();
    }
    
}
